'''Tests replicating example notebooks.'''
# pylint: disable=missing-function-docstring,missing-class-docstring

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from glob import glob
import os
from pathlib import Path
import subprocess

from ngautonml.algorithms.impl.algorithm_auto import FakeAlgorithmCatalogAuto
from ngautonml.problem_def.problem_def import ProblemDefinition
from ngautonml.tables.impl.table_auto import TableCatalogAuto
from ngautonml.wrangler.wrangler import Wrangler
# pylint: disable=duplicate-code
TableCatalogAuto()


def module_path() -> Path:
    return Path(__file__).parents[3]


def flowers_config():
    train_dir = module_path() / 'examples' / 'flowers' / 'train'
    test_dir = module_path() / 'examples' / 'flowers' / 'test'
    pdef = {
        "dataset": {
            "config": "image_tensor",
            "train_dir": train_dir,
            "test_dir": test_dir
        },
        "problem_type": {
            "data_type": "IMAGE",
            "task": "MULTICLASS_CLASSIFICATION"
        },
        "cross_validation": {
            "k": 1
        },
        "hyperparams": [
            {
                "_comments": [
                    "We need fewer epochs for a test environment without GPU."
                ],
                "select": {
                    "algorithm": "tf.keras.sequential"
                },
                "params": {
                    "epochs": {
                        "fixed": 2
                    }
                }
            }
        ]
    }
    return ProblemDefinition(pdef)


def test_wrangle_image() -> None:
    plugin_root = Path(__file__).parents[1]
    env = os.environ.copy()

    subprocess.run(['python', '-m', 'build'], cwd=plugin_root, check=True)

    pluginpaths = glob(str(plugin_root / 'dist' / 'ngautonml_image-*py3-none-any.whl'))
    assert len(pluginpaths) == 1, f'BUG: unexpected extra whl files: {pluginpaths}'
    pluginpath = Path(pluginpaths[0])
    subprocess.run(['pip', 'install', pluginpath], check=True, env=env)

    problem_definition = flowers_config()
    dut = Wrangler(
        problem_definition=problem_definition,
        algorithm_catalog=FakeAlgorithmCatalogAuto
    )
    got = dut.fit_predict_rank()

    assert list(got.train_results.values())[0].prediction.predictions_table.shape == (733, 1)

    subprocess.run(['pip', 'uninstall', '-y', pluginpath], check=False)



'''Class for Keras Sequential model'''
# pylint: disable=duplicate-code,too-many-arguments
from typing import Any, Dict, Iterable, List, Optional

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import numpy as np
import tensorflow as tf
import keras  # type: ignore[import-untyped]
from keras import layers  # type: ignore[import-untyped]
from ngautonml.algorithms.impl.fittable_algorithm_instance import (FittableAlgorithmInstance,
                                                                   UntrainedError)
from ngautonml.algorithms.impl.algorithm import Algorithm, MemoryAlgorithmCatalog
from ngautonml.catalog.catalog import upcast
from ngautonml.problem_def.task import DataType, TaskType
from ngautonml.wrangler.dataset import Dataset, DatasetKeys, TableFactory


class SequentialInstance(FittableAlgorithmInstance):
    '''This is an example image classifier using TensorFlow.

    https://www.tensorflow.org/tutorials/images/classification#compile_the_model
    '''
    _hyperparams: Dict[str, Any]
    _constructor: Any
    _class_names: List[str]
    _impl: Optional[keras.Sequential]

    def __init__(self, **overrides: Any):
        parent = overrides.pop('parent')
        self._constructor = keras.Sequential
        super().__init__(parent=parent)
        self._hyperparams = self.algorithm.hyperparams(**overrides)
        self._impl = None

    def hyperparams(self, **overrides) -> Dict[str, Any]:
        '''Report hyperparams with the ability to override them.'''
        retval = self._hyperparams.copy()
        retval.update(overrides)
        return retval

    def _mk_model(self, **overrides) -> keras.Sequential:
        params = self.algorithm.hyperparams(**overrides)
        print(params)
        return keras.Sequential([
            layers.Rescaling(1. / 255, input_shape=(params['img_height'], params['img_width'], 3)),
            layers.Conv2D(16, 3, padding='same', activation='relu'),
            layers.MaxPooling2D(),
            layers.Conv2D(32, 3, padding='same', activation='relu'),
            layers.MaxPooling2D(),
            layers.Conv2D(64, 3, padding='same', activation='relu'),
            layers.MaxPooling2D(),
            layers.Flatten(),
            layers.Dense(128, activation='relu'),
            layers.Dense(params['num_classes'])
        ])

    def fit(self, dataset: Optional[Dataset]) -> None:
        '''Fit the model.'''
        if dataset is None:
            return
        params = self.hyperparams()
        train_ds = dataset[DatasetKeys.KERAS_DS.value]
        assert isinstance(train_ds, tf.data.Dataset)
        val_ds = dataset[DatasetKeys.KERAS_VALIDATE.value]
        assert isinstance(val_ds, tf.data.Dataset)

        self._class_names = train_ds.class_names
        if params['num_classes'] is None:
            params['num_classes'] = len(self._class_names)
        self._impl = self._mk_model(**params)
        self._impl.compile(
            optimizer=self.hyperparams()['optimizer'],
            loss=keras.losses.SparseCategoricalCrossentropy(from_logits=True),
            metrics=['accuracy'])

        # Grab the epochs to pass to fit
        epochs = self.hyperparams()['epochs']

        # unsure why this was in a try/except
        self._impl.fit(train_ds,
                       validation_data=val_ds,
                       epochs=epochs)
        self._trained = True

    def predict(self, dataset: Optional[Dataset]) -> Optional[Dataset]:
        # TODO(Merritt): we should get train predictions somehow
        if not self._trained:
            raise UntrainedError(f'attempt to predict before fit for {self.catalog_name}')

        if dataset is None:
            return None

        assert self._impl is not None

        predictions = self._impl.predict(dataset[DatasetKeys.KERAS_DS.value])
        assert isinstance(predictions, np.ndarray)
        scores: Iterable[tf.Tensor] = tf.nn.softmax(predictions)
        prediction_result: List[str] = [self._class_names[np.argmax(score)] for score in scores]

        retval = dataset.output()

        assert dataset.metadata.target is not None

        retval.predictions_table = TableFactory({
            dataset.metadata.target.name: prediction_result
        })

        return retval


class KerasSequential(Algorithm):
    '''Class for Keras Sequential model'''
    _name = 'tf.keras.sequential'
    _tags = {
        'data_type': [DataType.IMAGE.name],
        'task': [TaskType.BINARY_CLASSIFICATION.name,
                 TaskType.MULTICLASS_CLASSIFICATION.name],
        'for_tests': ['true']
    }
    _default_hyperparams = {
        'img_height': 128,
        'img_width': 128,
        'num_classes': None,
        'optimizer': 'adam',
        'epochs': 2,
    }
    _instance_constructor = SequentialInstance


def register(catalog: MemoryAlgorithmCatalog, *args, **kwargs) -> None:
    '''Register all the objects in this file.'''
    model = KerasSequential(*args, **kwargs)
    catalog.register(model, model.name, upcast(model.tags))



'''NGAutonML plugin for image classification.

Currently (10/25/2023) implemented using keras.'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from pathlib import Path

from ngautonml.catalog.catalog import Catalog
from ngautonml.catalog.pathed_catalog import PathedCatalog


def templates_make_catalog(**kwargs) -> Catalog:
    '''Make the plugin templates catalog.'''
    return PathedCatalog([Path(__file__).parent / 'templates'], **kwargs)


def algorithms_make_catalog(**kwargs) -> Catalog:
    '''Make the plugin algorithms catalog.'''
    return PathedCatalog([Path(__file__).parent / 'algorithms'], **kwargs)


def data_loaders_make_catalog(**kwargs) -> Catalog:
    '''Make the plugin data_loaders catalog.'''
    return PathedCatalog([Path(__file__).parent / 'data_loaders'], **kwargs)


def splitters_make_catalog(**kwargs) -> Catalog:
    '''Make the plugin splitters catalog.'''
    return PathedCatalog([Path(__file__).parent / 'splitters'], **kwargs)



'''Pipeline template for image classification.'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from typing import Dict, List, Optional

from ngautonml.algorithms.impl.algorithm import AlgorithmCatalog
from ngautonml.problem_def.task import DataType, TaskType
from ngautonml.templates.impl.pipeline_step import GeneratorInterface
from ngautonml.templates.impl.pipeline_template import PipelineTemplate
from ngautonml.templates.impl.template import TemplateCatalog


class ImageClassificationTemplate(PipelineTemplate):
    '''Pipeline template for image classification.'''
    _name = 'image_classifier'
    _tags: Dict[str, List[str]] = {
        'task': [TaskType.BINARY_CLASSIFICATION.name,
                 TaskType.MULTICLASS_CLASSIFICATION.name],
        'data_type': [DataType.IMAGE.name],
    }

    def __init__(self,
                 name: Optional[str] = None,
                 tags: Optional[Dict[str, List[str]]] = None,
                 algorithm_catalog: Optional[AlgorithmCatalog] = None,
                 generator: Optional[GeneratorInterface] = None):
        super().__init__(name=name or self._name,
                         tags=tags or self._tags,
                         algorithm_catalog=algorithm_catalog, generator=generator)
        self.query(
            task=[TaskType.BINARY_CLASSIFICATION.name,
                  TaskType.MULTICLASS_CLASSIFICATION.name],
            data_type=DataType.IMAGE.name).set_name('image_classifier')


def register(catalog: TemplateCatalog,
             algorithm_catalog: Optional[AlgorithmCatalog] = None,
             generator: Optional[GeneratorInterface] = None):
    '''Register all the objects in this file.'''
    catalog.register(
        ImageClassificationTemplate(
            algorithm_catalog=algorithm_catalog,
            generator=generator))



'''Tests for image_tensor_params.py.'''
# mypy: disable-error-code="attr-defined"

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from .image_tensor_params import ImageTensorParams

# pylint: disable=missing-function-docstring,duplicate-code


def test_basic_params() -> None:
    clause = {
        'train_dir': '/some/path'
    }

    dut = ImageTensorParams(clause=clause)

    assert dut.train_dir == '/some/path'


def test_defaullted_param() -> None:
    clause = {
        'train_dir': '/some/path'
    }

    dut = ImageTensorParams(clause=clause)

    assert dut.batch_size == 32


def test_chrismas_tree() -> None:
    clause = {
        'train_dir': '/some/train/path',
        'test_dir': '/some/test/path',
        'validation_split': 0.9,
        'seed': 1701,
        'img_height': 1024,
        'img_width': 1024,
        'batch_size': 10,
    }

    dut = ImageTensorParams(clause=clause)

    assert dut.train_dir == '/some/train/path'
    assert dut.test_dir == '/some/test/path'
    assert dut.validation_split == 0.9
    assert dut.seed == 1701
    assert dut.img_height == 1024
    assert dut.img_width == 1024
    assert dut.batch_size == 10



'''Nop splitter that puts the entire dataset in TRAIN, and VALIDATE, and GROUND_TRUTH.'''

import numpy as np
import pandas as pd

from ngautonml.problem_def.cross_validation_config import CrossValidationConfig
from ngautonml.problem_def.task import DataType, TaskType
from ngautonml.splitters.impl.splitter import Splitter, SplitDataset, Fold, SplitterCatalog
from ngautonml.wrangler.dataset import Dataset, DatasetKeys, TableFactory


class NopSplitter(Splitter):
    '''Splitter that passes a dataset through unaltered.
    '''
    _name = 'nop'
    _tags = {
        'task': [
            TaskType.BINARY_CLASSIFICATION.name,
            TaskType.MULTICLASS_CLASSIFICATION.name
        ],
        'data_type': [DataType.IMAGE.name],
    }

    def split(self,
              dataset: Dataset,
              **unused_overrides) -> SplitDataset:
        '''Create a trivial SplitDataset with a single fold containing the entire dataset.

        A copy of the dataset gets put in TRAIN and VALIDATE for that fold,
        and a copy gets put in GROUND_TRUTH.
        '''

        val_ds = dataset[DatasetKeys.KERAS_VALIDATE.value]
        labels = np.concatenate([y for _, y in val_ds], axis=0)
        label_names = val_ds.class_names
        ground_truth_list = [label_names[label] for label in labels]
        assert dataset.metadata.target is not None
        ground_truth_df = pd.DataFrame(
            {dataset.metadata.target.name: ground_truth_list}
        )
        ground_truth_dataset = dataset.output()
        ground_truth_dataset.ground_truth_table = TableFactory(ground_truth_df)
        validate = dataset.output()
        validate[DatasetKeys.KERAS_DS.value] = val_ds
        return SplitDataset([Fold(train=dataset, validate=validate)])


def register(catalog: SplitterCatalog, *args, cv_config: CrossValidationConfig, **kwargs):
    '''Register all the objects in this file.'''
    catalog.register(NopSplitter(*args, cv_config=cv_config, **kwargs))
